from __future__ import annotations

import wrapt

import numpy as np
from matplotlib.axes import Axes
from matplotlib.container import BarContainer
from matplotlib.patches import Polygon

from maidr.core.context_manager import ContextManager
from maidr.core.enum import PlotType
from maidr.core.figure_manager import FigureManager
from maidr.patch.common import common


@wrapt.patch_function_wrapper(Axes, "hist")
def mpl_hist(
    wrapped, _, args, kwargs
) -> tuple[
    np.ndarray | list[np.ndarray],
    np.ndarray,
    BarContainer | Polygon | list[BarContainer | Polygon],
]:
    # Don't proceed if the call is made internally by the patched function.
    if ContextManager.is_internal_context():
        return wrapped(*args, **kwargs)

    # Set the internal context to avoid cyclic processing.
    with ContextManager.set_internal_context():
        # Patch `ax.hist()`.
        n, bins, plot = wrapped(*args, **kwargs)

    # Extract the histogram data points for MAIDR from the plots.
    ax = FigureManager.get_axes(plot)
    FigureManager.create_maidr(ax, PlotType.HIST)

    # Return to the caller.
    return n, bins, plot


@wrapt.patch_function_wrapper("seaborn", "histplot")
def sns_hist(wrapped, instance, args, kwargs) -> Axes:
    return common(PlotType.HIST, wrapped, instance, args, kwargs)



from __future__ import annotations
from typing import Literal

import io
import json
import uuid

from htmltools import HTML, HTMLDocument, RenderedHTML, tags, Tag
from lxml import etree

from matplotlib.figure import Figure

from maidr.core.plot import MaidrPlot


class Maidr:
    """
    A class to handle the rendering and interaction of matplotlib figures with additional metadata.

    Attributes
    ----------
    _fig : Figure
        The matplotlib figure associated with this instance.
    _plots : list[MaidrPlot]
        A list of MaidrPlot objects which hold additional plot-specific configurations.

    Methods
    -------
    render(lib_prefix=None, include_version=True)
        Creates and returns a rendered HTML representation of the figure.
    save_html(file, lib_dir=None, include_version=True)
        Saves the rendered HTML representation to a file.
    show(renderer='auto')
        Displays the rendered HTML content in the specified rendering context.
    """

    def __init__(self, fig: Figure) -> None:
        """Create a new Maidr for the given ``matplotlib.figure.Figure``."""
        self._fig = fig
        self._plots = list()

    @property
    def fig(self) -> Figure:
        """Return the ``matplotlib.figure.Figure`` associated with this object."""
        return self._fig

    @property
    def plots(self) -> list[MaidrPlot]:
        """Return the list of plots extracted from the ``fig``."""
        return self._plots

    def render(
        self, *, lib_prefix: str | None = "lib", include_version: bool = True
    ) -> RenderedHTML:
        """
        Render the document.

        Parameters
        ----------
        lib_prefix : str, default="lib"
            A prefix to add to relative paths to dependency files.
        include_version : bool, default=True
            Whether to include the version number in the dependency's folder name.
        """
        html = self._create_html_doc()
        return html.render(lib_prefix=lib_prefix, include_version=include_version)

    def save_html(
        self, file: str, *, lib_dir: str | None = "lib", include_version: bool = True
    ) -> str:
        """
        Save the HTML representation of the figure with MAIDR to a file.

        Parameters
        ----------
        file : str
            The file to save to.
        lib_dir : str, default="lib"
            The directory to save the dependencies to (relative to the file's directory).
        include_version : bool, default=True
            Whether to include the version number in the dependency folder name.
        """
        html = self._create_html_doc()
        return html.save_html(file, libdir=lib_dir, include_version=include_version)

    def show(self, renderer: Literal["auto", "ipython", "browser"] = "auto") -> object:
        """
        Preview the HTML content using the specified renderer.

        Parameters
        ----------
        renderer : Literal["auto", "ipython", "browser"], default="auto"
            The renderer to use for the HTML preview.
        """
        html = self._create_html_tag()
        return html.show(renderer)

    def _create_html_tag(self) -> Tag:
        """Create the MAIDR HTML using HTML tags."""
        svg = self._get_svg()
        maidr = f"\nlet maidr = {json.dumps(self._flatten_maidr(), indent=2)}\n"

        # Inject plot's svg and MAIDR structure into html tag.
        return Maidr._inject_plot(svg, maidr)

    def _create_html_doc(self) -> HTMLDocument:
        """Create an HTML document from Tag objects."""
        return HTMLDocument(self._create_html_tag(), lang="en")

    def _flatten_maidr(self) -> dict | list[dict]:
        """Return a single plot schema or a list of schemas from the Maidr instance."""
        maidr = [plot.schema for plot in self._plots]
        return maidr if len(maidr) != 1 else maidr[0]

    def _get_svg(self) -> HTML:
        """Extract the chart SVG from ``matplotlib.figure.Figure``."""
        svg_buffer = io.StringIO()
        self._fig.savefig(svg_buffer, format="svg")
        str_svg = svg_buffer.getvalue()

        etree.register_namespace("svg", "http://www.w3.org/2000/svg")
        tree_svg = etree.fromstring(str_svg.encode(), parser=None)
        root_svg = None
        # Find the `svg` tag and set unique id if not present else use it.
        for element in tree_svg.iter(tag="{http://www.w3.org/2000/svg}svg"):
            if "id" not in element.attrib:
                element.attrib["id"] = Maidr._unique_id()
            root_svg = element
            self._set_maidr_id(element.attrib["id"])
            break

        svg_buffer = io.StringIO()  # Reset the buffer
        svg_buffer.write(
            etree.tostring(
                root_svg, pretty_print=True, encoding="unicode"  # type: ignore
            )
        )

        return HTML(svg_buffer.getvalue())

    def _set_maidr_id(self, maidr_id: str) -> None:
        """Set a unique identifier to each ``MaidrPlot``."""
        for maidr in self._plots:
            maidr.set_id(maidr_id)

    @staticmethod
    def _unique_id() -> str:
        """Generate a unique identifier string using UUID4."""
        return str(uuid.uuid4())

    @staticmethod
    def _inject_plot(plot: HTML, maidr: str) -> Tag:
        """Embed the plot and associated MAIDR scripts into the HTML structure."""
        return tags.html(
            tags.head(
                tags.meta(charset="UTF-8"),
                tags.title("MAIDR"),
                tags.link(
                    rel="stylesheet",
                    href="https://cdn.jsdelivr.net/npm/maidr/dist/maidr_style.min.css",
                ),
                tags.script(
                    type="text/javascript",
                    src="https://cdn.jsdelivr.net/npm/maidr/dist/maidr.min.js",
                ),
            ),
            tags.body(
                tags.div(plot),
            ),
            tags.script(maidr),
        )



from __future__ import annotations

from typing import Literal, Any

from matplotlib.axes import Axes
from matplotlib.container import BarContainer

from maidr.core import Maidr
from maidr.core.enum import PlotType
from maidr.core.figure_manager import FigureManager


def show(plot: Any, renderer: Literal["auto", "ipython", "browser"] = "auto") -> object:
    ax = FigureManager.get_axes(plot)
    maidr = FigureManager.get_maidr(ax.get_figure())
    return maidr.show(renderer)


def save_html(
    plot: Any, file: str, *, lib_dir: str | None = "lib", include_version: bool = True
) -> str:
    ax = FigureManager.get_axes(plot)
    maidr = FigureManager.get_maidr(ax.get_figure())
    return maidr.save_html(file, lib_dir=lib_dir, include_version=include_version)


def stacked(plot: Axes | BarContainer) -> Maidr:
    ax = FigureManager.get_axes(plot)
    return FigureManager.create_maidr(ax, PlotType.STACKED)


def close() -> None:
    pass


def test_enum_comparison():
    print(
        f"PlotType.COUNT == PlotType.BAR: {PlotType.COUNT == PlotType.BAR}"
    )  # Should be False
    print(
        f"PlotType.COUNT is PlotType.BAR: {PlotType.COUNT is PlotType.BAR}"
    )  # Should be False
    print(
        f"PlotType.COUNT == PlotType.COUNT: {PlotType.COUNT == PlotType.COUNT}"
    )  # Should be True
    print(
        f"PlotType.BAR == PlotType.BAR: {PlotType.BAR == PlotType.BAR}"
    )  # Should be True
    print(
        f"PlotType.COUNT is PlotType.COUNT: {PlotType.COUNT is PlotType.COUNT}"
    )  # Should be True
    print(f"PlotType.BAR is PlotType.BAR: {PlotType.BAR is PlotType.BAR}")



__version__ = "0.4.1"

from .core import Maidr
from .core.enum import PlotType
from .patch import barplot, boxplot, heatmap, histogram, lineplot, scatterplot
from .maidr import save_html, show, stacked

__all__ = [
    "save_html",
    "show",
    "stacked",
]



from __future__ import annotations

from matplotlib.axes import Axes
from matplotlib.lines import Line2D

from maidr.core.enum import MaidrKey, PlotType
from maidr.core.plot import MaidrPlot
from maidr.exception import ExtractionError
from maidr.utils.mixin import LineExtractorMixin


class LinePlot(MaidrPlot, LineExtractorMixin):
    def __init__(self, ax: Axes) -> None:
        super().__init__(ax, PlotType.LINE)

    def _extract_plot_data(self) -> list[dict]:
        plot = self.extract_line(self.ax)
        data = LinePlot._extract_line_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    @staticmethod
    def _extract_line_data(plot: Line2D | None) -> list[dict] | None:
        if plot is None or plot.get_xydata() is None:
            return None

        return [
            {
                MaidrKey.X.value: float(x),
                MaidrKey.Y.value: float(y),
            }
            for x, y in plot.get_xydata()
        ]



from __future__ import annotations

from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.collections import QuadMesh

from maidr.core.enum import MaidrKey, PlotType
from maidr.core.plot import MaidrPlot
from maidr.exception import ExtractionError
from maidr.utils.mixin import (
    DictMergerMixin,
    LevelExtractorMixin,
    ScalarMappableExtractorMixin,
)


class HeatPlot(
    MaidrPlot, LevelExtractorMixin, ScalarMappableExtractorMixin, DictMergerMixin
):
    def __init__(self, ax: Axes, **kwargs) -> None:
        self._fill_label = kwargs.pop("fill_label")
        super().__init__(ax, PlotType.HEAT)

    def render(self) -> dict:
        base_maidr = super().render()
        heat_maidr = {
            MaidrKey.LABELS: {
                MaidrKey.FILL: self._fill_label,
            },
        }
        return self.merge_dict(base_maidr, heat_maidr)

    def _extract_axes_data(self) -> dict:
        base_ax_schema = super()._extract_axes_data()
        heat_ax_schema = {
            MaidrKey.X: {
                MaidrKey.LEVEL: self.extract_level(self.ax, MaidrKey.X),
            },
            MaidrKey.Y: {
                MaidrKey.LEVEL: self.extract_level(self.ax, MaidrKey.Y),
            },
        }
        return self.merge_dict(base_ax_schema, heat_ax_schema)

    def _extract_plot_data(self) -> list[list]:
        plot = self.extract_scalar_mappable(self.ax)
        data = HeatPlot._extract_scalar_mappable_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    @staticmethod
    def _extract_scalar_mappable_data(sm: ScalarMappable | None) -> list[list] | None:
        if sm is None or sm.get_array() is None:
            return None

        array = sm.get_array().data
        if isinstance(sm, QuadMesh):
            # Data gets flattened in ScalarMappable, when the plot is from QuadMesh.
            # So, reshaping the data to reflect the original quadrilaterals
            m, n, _ = sm.get_coordinates().shape
            array = array.reshape(m - 1, n - 1)  # Coordinates shape is (M + 1, N + 1)

        return [list(map(float, row)) for row in array]



from enum import Enum


class PlotType(str, Enum):
    """An enumeration of plot types supported by MAIDR."""

    BAR = "bar"
    BOX = "box"
    COUNT = "count"
    DODGED = "dodged_bar"
    HEAT = "heat"
    HIST = "hist"
    LINE = "line"
    SCATTER = "point"
    STACKED = "stacked_bar"



import contextlib
import contextvars
import threading

import wrapt


class ContextManager:
    _instance = None
    _lock = threading.Lock()

    _internal_context = contextvars.ContextVar("internal_context", default=False)

    def __new__(cls):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super(ContextManager, cls).__new__()
        return cls._instance

    @classmethod
    @contextlib.contextmanager
    def set_internal_context(cls):
        token_internal_context = cls._internal_context.set(True)
        try:
            yield
        finally:
            cls._internal_context.reset(token_internal_context)

    @classmethod
    def is_internal_context(cls):
        return cls._internal_context.get()


@wrapt.decorator
def manage_context(wrapped=None, _=None, args=None, kwargs=None):
    # Don't proceed if the call is made internally by the patched function.
    if ContextManager.is_internal_context():
        return wrapped(*args, **kwargs)

    # Set the internal context to avoid cyclic processing.
    with ContextManager.set_internal_context():
        return wrapped(*args, **kwargs)



from typing import Any

from maidr.core.enum import PlotType


class ExtractionError(Exception):
    """
    Exception raised for errors encountered during the MAIDR data extraction process.

    Parameters
    ----------
    plot_type : PlotType
        The type of plot from which data extraction failed.
    plot : Any
        The plot object that caused the extraction error.
    """

    def __init__(self, plot_type: PlotType, plot: Any):
        self.message = (
            f"Error extracting data for {plot_type.value} plot type from {type(plot)}."
        )
        super().__init__(self.message)
